# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import jwt
import logging
import time

from leihsldap.config import config

# Logger
logger = logging.getLogger(__name__)


def token_data(token: str) -> tuple[dict[str, list], str, str, bool]:
    '''Verify JWT token and extract data contained within token.

    :param token: The JWT token.
    :returns: Tuple of original data, email, login and if the user is already
        registered
    '''
    options = {'verify_exp': not config('token', 'allow_expired')}
    private_key = config('token', 'private_key')
    data = jwt.decode(token, private_key, ['ES256'], options)

    email = data.get('email')
    login = data.get('login')
    if type(email) is not str:
        raise RuntimeError('Field email must be set.')
    # If a user does not exist, Leihs will treat the entered value as an email.
    # If the value identifies a user, Leihs will supply the login it knows.
    # We always set login, so we can use this to check if a user is registered.
    if type(login) is str:
        logger.debug('Login attempt from already registered user `%s` (`%s`)',
                     login, email)
        return data, email, login, True
    login = email.split('@', 1)[0]
    logger.debug('Login attempt from unregistered user `%s` (`%s`)',
                 login, email)
    return data, email, login, False


def response_url(token: str, data: dict[str, str]) -> str:
    '''Generate response URL to redirect to after authentication.
    The URL should point to Leihs and will contain a success token with user
    information to identify the user in Leihs.

    :param token: Original request token from Leihs.
    :param data: User data to pass on to Leihs.
    '''
    # generate success token
    iat = int(time.time())
    exp = iat + config('token', 'validity')
    private_key = config('token', 'private_key')
    success_token = jwt.encode({
            'sign_in_request_token': token,
            'email': data.get('email'),
            'iat': iat,
            'exp': exp,
            'success': True
            }, private_key, 'ES256')

    logger.debug('returning success token: %s', success_token)
    base_url = data['server_base_url']
    path = data['path']
    return f'{base_url}{path}?token={success_token}'



# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import requests

from typing import Optional

from leihsldap.config import config

# Logger
logger = logging.getLogger(__name__)


def api(method: str, path: str, **kwargs) -> requests.models.Response:
    '''Execute an HTTP request against the Leihs API.
    This uses the API token from the configuration file.

    :param method: HTTP method to use
    :param path: Path of the request URL
    :returns: HTTP response
    '''
    base_url = config('leihs', 'url', allow_empty=False)
    url = f'{base_url}{path}'
    headers = {'Accept': 'application/json',
               'Content-Type': 'application/json',
               'Authorization': 'Token ' + config('leihs', 'api_token')}
    logger.debug('Sending request to %s', url)
    return requests.request(method, url, headers=headers, **kwargs)


def check(response: requests.models.Response, error_message: str) -> None:
    '''Check if the response contains an HTTP error code and raise an exception
    if it does. The exception message will contain additional information
    returned by the request.

    :param response: HTTP response return by requests library
    :param error_message: Error message to include if an exception is raised
    :raises RuntimeError: HTTP error response detected
    '''
    if response.status_code >= 300:
        message = '\n'.join([
            error_message,
            f'Status code: {response.status_code}',
            response.text
            ])
        raise RuntimeError(message)


def register_user(email: str,
                  firstname: Optional[str] = None,
                  lastname: Optional[str] = None,
                  username: Optional[str] = None,
                  groups: list[str] = []):
    '''Register a new user with Leihs.
    Skip registration if the user already exists.

    :param email: Email address
    :param firstname: The user's given name
    :param lastname: The user's family name
    :param username: The user's login
    :param groups: List of groups to add user to when initially created
    '''
    # register new user
    user_data = {
        'email': email,
        'firstname': firstname,
        'lastname': lastname,
        'account_enabled': True,
        'password_sign_in_enabled': False,
        'login': username,
        'extended_info': None
        }
    logger.debug('Trying to register user: %s', user_data)
    response = api('post', '/admin/users/', json=user_data)

    # It's fine if we get a conflict.
    # That just means, the user is already registered
    if response.status_code != 409:
        logger.debug('New user created. Adding authentication and groups.')
        check(response, 'Could not create user')

        # add the newly created user to the authentication system
        user_data = response.json()
        add_user_to_auth(user_data['id'])

        # add user to groups
        for group in groups:
            group_data = create_group(group)
            add_user_to_group(user_data['id'], group_data['id'])


def add_user_to_auth(user_id: str) -> None:
    '''Add user to the authentication system.

    :param user_id: Identifier of the user to add.
    '''
    auth_id = config('auth-system', 'id')
    path = f'/admin/system/authentication-systems/{auth_id}/users/{user_id}'
    logger.debug('Adding user `%s` to authentication system `%s`',
                 user_id, auth_id)
    response = api('put', path)
    check(response, 'Could not add user to authentication system')


def register_auth_system() -> None:
    '''Register the authentication system with Leihs.
    Registration data are taken from the configuration.

    :returns: Dictionary with authentication system data
    '''
    # register system
    system_data = {
        'description': config('auth-system', 'description'),
        'enabled': True,
        'external_public_key': config('token', 'public_key'),
        'external_sign_in_url': config('auth-system', 'url'),
        'id': config('auth-system', 'id', allow_empty=False),
        'internal_private_key': config('token', 'private_key'),
        'internal_public_key': config('token', 'public_key'),
        'name': config('auth-system', 'name'),
        'priority': config('auth-system', 'priority') or 3,
        'send_email': True,
        'send_login': True,
        'type': 'external',
        'sign_up_email_match': config('auth-system', 'email_match')
        }
    logger.debug('Trying to register authentication system `%s´',
                 system_data['id'])
    response = api('post',
                   '/admin/system/authentication-systems/',
                   json=system_data)

    # If we got a 409, the system is already registered and everything is good
    if response.status_code == 409:
        logger.debug('Authentication system was already registered.')
        return

    # If we got anything else, check for errors
    check(response, 'Could not register authentication system')


def create_group(name: str) -> dict:
    '''Create group if it does not yet exists.

    :param name: Name of the group to create. Also used as org_id.
    :returns: Dictionary of user data
    '''
    group_data = {
        'name': name,
        'org_id': name,
        'organization': 'leihs-local'
        }
    logger.debug('Trying to create group %s', name)
    response = api('post', '/admin/groups/', json=group_data)

    # If we just created the group, we have all data we need
    if response.status_code != 409:
        check(response, 'Could not create group')
        return response.json()

    # if the group already existed, get the existing group's data
    logger.debug('Group already existed. Getting info from existing group')
    del group_data['name']
    response = api('get', '/admin/groups/', params=group_data)
    check(response, 'Could not get group data')
    groups_found = response.json().get('groups', [])
    if len(groups_found) != 1:
        raise RuntimeError(f'Got invalid group data: {groups_found}')
    return groups_found[0]


def add_user_to_group(user_id: str, group_id: str):
    '''Add a user to a group in Leihs.

    :param user_id: Identifier of the user to add
    :param group_id: Identifier of the Group to add the user to
    '''
    logger.debug('Trying to add user `%s` to group `%s`.', user_id, group_id)
    response = api('put', f'/admin/groups/{group_id}/users/{user_id}')
    check(response, 'Could not add user to group')
    return response



# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

from ldap3 import Server, Connection, ALL, AUTO_BIND_NO_TLS

from leihsldap.config import config

# Logger
logger = logging.getLogger(__name__)


def ldap_login(username: str, password: str) -> dict[str, list]:
    '''Login to LDAP and return user attributes.

    The idea of this is basically for the user to login to LDAP and request its
    own attributes. Obviously, this code will do that for the user with the
    provided credentials.

    :param username: Username to log in with.
    :param password: Password to log in with.
    :returns: Dictionary containing requested user attributes.
    '''
    user_dn = config('ldap', 'user_dn').format(username=username)

    logger.debug('Trying to log into LDAP with user_dn `%s`', user_dn)
    server = Server(config('ldap', 'server'),
                    port=config('ldap', 'port'),
                    use_ssl=True,
                    get_info=ALL)
    # Note: AUTO_BIND_NO_TLS means no Start TLS
    # See: https://github.com/cannatag/ldap3/issues/1061
    conn = Connection(server, user_dn, password, auto_bind=AUTO_BIND_NO_TLS)
    logger.debug('Login successful with user_dn `%s`', user_dn)

    attributes = list(filter(bool, [
        config('ldap', 'userdata', 'email', 'field'),
        config('ldap', 'userdata', 'name', 'family'),
        config('ldap', 'userdata', 'name', 'given')]))
    attributes += config('ldap', 'userdata', 'groups', 'fields') or []

    logger.debug('Searching for user data')
    conn.search(
            config('ldap', 'base_dn'),
            config('ldap', 'search_filter').format(username=username),
            attributes=attributes)
    if len(conn.entries) != 1:
        raise ValueError('Search must return exactly one result', conn.entries)
    logger.debug('Found user data')
    return conn.entries[0].entry_attributes_as_dict



# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import yaml

from flask import Flask, request, redirect, render_template
from functools import wraps
from jwt.exceptions import DecodeError, ExpiredSignatureError
from ldap3.core.exceptions import LDAPBindError, LDAPPasswordIsMandatoryError

from leihsldap.authenticator import response_url, token_data
from leihsldap.config import config
from leihsldap.ldap import ldap_login
from leihsldap.leihs_api import register_user, register_auth_system

# Logger
logger = logging.getLogger(__name__)

flask_config = {}
if config('ui', 'directories', 'template'):
    flask_config['template_folder'] = config('ui', 'directories', 'template')
if config('ui', 'directories', 'static'):
    flask_config['static_folder'] = config('ui', 'directories', 'static')
app = Flask(__name__, **flask_config)

__error = {}


def error(error_id: str, code: int) -> tuple[str, int]:
    '''Generate error page based on data defined in `error.yml` and the given
    error identifier.

    :param error_id: String identifying the error to render.
    :param code: HTTP status code to return.
    :returns: Tuple of data for Flask response
    '''
    if not __error:
        with open(os.path.dirname(__file__) + '/error.yml', 'r') as f:
            globals()['__error'] = yaml.safe_load(f)
    error_data = __error[error_id].copy()
    error_data['leihs_url'] = config('leihs', 'url')
    return render_template('error.html', **error_data), code


def handle_errors(function):
    '''Decorator handling common errors.
    This will cause an errpr page to be rendered.

    :param function: Function to wrap.
    '''
    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except DecodeError as e:
            logger.info('Error decoding token: %s', e)
            return error('invalid_token', 400)
        except ExpiredSignatureError as e:
            logger.info('Token expired: %s', e)
            return error('expired_token', 400)
        except (LDAPBindError, LDAPPasswordIsMandatoryError) as e:
            logger.info('LDAP login failed: %s', e)
            return error('invalid_credentials', 403)
    return wrapper


@app.before_first_request
def before_first_request():
    '''Try registering the authentication system.
    '''
    logger.info('Trying to register authentication system')
    register_auth_system()


@app.errorhandler(500)
def internal_server_error(e):
    '''Handle internal server errors.
    This causes the app to render an error page similar to known and caught
    errors.
    '''
    return error('internal', 500)


@app.route('/', methods=['GET'])
@handle_errors
def login_page():
    '''Render login page.
    This is the page users will end up on when redirected from Leihs.
    '''
    token = request.args.get('token')
    if not token:
        logger.debug('No token provided')
        return error('no_token', 400)
    _, email, user, _ = token_data(token)
    return render_template('login.html', token=token, user=user)


@app.route('/', methods=['POST'])
@handle_errors
def login():
    '''Handle login POST requests.
    The POST request form data must contain the fields:

    - token: JWT token received from and signed by Leihs
    - password: The password the user tries to sign in with
    '''
    # get form data
    token = request.form.get('token')
    password = request.form.get('password')

    # verify token and get login data
    data, email, user, registered = token_data(token)

    # Login to and get user data from LDAP
    user_data = ldap_login(user, password)

    # Get list of groups the user should be in
    group_fields = config('ldap', 'userdata', 'groups', 'fields') or []
    groups = [group for field in group_fields for group in user_data[field]]

    # Check if to fall back to the LDAP email address
    email_overwrite = config('ldap', 'userdata', 'email', 'overwrite')
    email_fallback = config('ldap', 'userdata', 'email', 'fallback')
    email_invalid = not email or '@' not in email
    if email_overwrite or email_fallback and email_invalid:
        email_field = config('ldap', 'userdata', 'email', 'field')
        email = user_data[email_field][0]
        data['email'] = email

    # Make sure user is registered with Leihs
    register_user(
            email,
            firstname=user_data['givenName'][0],
            lastname=user_data['sn'][0],
            username=user,
            groups=groups)

    # Redirect back to Leihs with success token
    return redirect(response_url(token, data), code=302)



# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''
Load and handle Leihs LDAP Authenticator configuration.
'''

import logging
import os
import yaml

# Logger
logger = logging.getLogger(__name__)

__config = {}


def configuration_file():
    '''Find the best match for the configuration file.  The configuration file
    locations taken into consideration are (in this particular order):

    - ``./leihs-ldap.yml``
    - ``~/leihs-ldap.yml``
    - ``/etc/leihs-ldap.yml``

    :return: configuration file name or None
    '''
    if os.path.isfile('./leihs-ldap.yml'):
        return './leihs-ldap.yml'
    expanded_file = os.path.expanduser('~/leihs-ldap.yml')
    if os.path.isfile(expanded_file):
        return expanded_file
    if os.path.isfile('/etc/leihs-ldap.yml'):
        return '/etc/leihs-ldap.yml'


def update_configuration(filename=None):
    '''Update configuration.
    '''
    cfgfile = filename or configuration_file()
    if not cfgfile:
        return {}
    print(f'Loading configuration from {cfgfile}')
    with open(cfgfile, 'r') as f:
        cfg = yaml.safe_load(f)
    globals()['__config'] = cfg

    # update logger
    loglevel = cfg.get('loglevel', 'INFO').upper()
    logging.root.setLevel(loglevel)
    logger.info('Log level set to %s', loglevel)

    return cfg


def config(*args, allow_empty=True):
    '''Get a specific configuration value or the whole configuration, loading
    the configuration file if it was not before.

    :param key: optional configuration key to return
    :type key: string
    :return: dictionary containing the configuration or configuration value
    '''
    cfg = __config or update_configuration()
    for key in args:
        if cfg is None:
            if allow_empty:
                return
            raise KeyError(f'Missing configuration key {args}')
        cfg = cfg.get(key)
    return cfg



from setuptools import setup, find_packages
import os

path = os.path.abspath(os.path.dirname(__file__))


def read(filename):
    with open(os.path.join(path, filename), encoding='utf-8') as f:
        return f.read()


setup(
    name='leihs-ldap-authenticator',
    author='Lars Kiesow <kiesow@elan-ev.de>',
    license='GPLv3+',
    version='0.2',
    url='https://github.com/elan-ev/leihs-ldap-authenticator',
    packages=find_packages(),
    install_requires=read('requirements.txt').split(),
    include_package_data=True,
    long_description=read('README.md'),
    long_description_content_type='text/markdown'
)



# LDAP based authentication handler for Leihs
# Copyright 2022 ELAN e.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from leihsldap.config import update_configuration


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='LDAP based authentication handler for Leihs',
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        '-c', '--config',
        type=str,
        default=None,
        help='Path to a configuration file'
    )
    args = parser.parse_args()
    if args.config:
        update_configuration(args.config)

    # Since `app` will use the configuration,
    # load it only after we updated the configuration location
    from leihsldap.web import app
    app.run()
